@@ -1,24 +1,31 @@
-from six.moves import BaseHTTPServer
+# mypy: allow-untyped-defs
+
 import errno
+import http.server
 import os
 import socket
-from six.moves.socketserver import ThreadingMixIn
+from socketserver import ThreadingMixIn
 import ssl
 import sys
 import threading
 import time
 import traceback
-from six import binary_type, text_type
 import uuid
 from collections import OrderedDict
-
-from six.moves.queue import Queue
+from queue import Empty, Queue
+from typing import Dict
 
 from h2.config import H2Configuration
 from h2.connection import H2Connection
 from h2.events import RequestReceived, ConnectionTerminated, DataReceived, StreamReset, StreamEnded
+from h2.exceptions import StreamClosedError, ProtocolError
+from h2.settings import SettingCodes
+from h2.utilities import extract_method_header
 
-from six.moves.urllib.parse import urlsplit, urlunsplit
+from urllib.parse import urlsplit, urlunsplit
+
+from mod_pywebsocket import dispatch
+from mod_pywebsocket.handshake import HandshakeException, AbortedByUserException
 
 from . import routes as default_routes
 from .config import ConfigBuilder
@@ -26,10 +33,21 @@ from .logger import get_logger
 from .request import Server, Request, H2Request
 from .response import Response, H2Response
 from .router import Router
-from .utils import HTTPException
+from .utils import HTTPException, isomorphic_decode, isomorphic_encode
 from .constants import h2_headers
+from .ws_h2_handshake import WsH2Handshaker
 
-"""HTTP server designed for testing purposes.
+# We need to stress test that browsers can send/receive many headers (there is
+# no specified limit), but the Python stdlib has an arbitrary limit of 100
+# headers. Hitting the limit leads to HTTP 431, so we monkey patch it higher.
+# https://bugs.python.org/issue26586
+# https://github.com/web-platform-tests/wpt/pull/24451
+import http.client
+assert isinstance(getattr(http.client, '_MAXHEADERS'), int)
+setattr(http.client, '_MAXHEADERS', 512)
+
+"""
+HTTP server designed for testing purposes.
 
 The server is designed to provide flexibility in the way that
 requests are handled, and to provide control both of exactly
@@ -63,7 +81,7 @@ handler returns, or for directly writing to the output stream.
 """
 
 
-class RequestRewriter(object):
+class RequestRewriter:
     def __init__(self, rules):
         """Object for rewriting the request path.
 
@@ -88,7 +106,7 @@ class RequestRewriter(object):
         :param output_path: Path to replace the input path with in
                             the request.
         """
-        if isinstance(methods, (binary_type, text_type)):
+        if isinstance(methods, (bytes, str)):
             methods = [methods]
         self.rules[input_path] = (methods, output_path)
 
@@ -111,7 +129,7 @@ class RequestRewriter(object):
                 request_handler.path = new_url
 
 
-class WebTestServer(ThreadingMixIn, BaseHTTPServer.HTTPServer):
+class WebTestServer(ThreadingMixIn, http.server.HTTPServer):
     allow_reuse_address = True
     acceptable_errors = (errno.EPIPE, errno.ECONNABORTED)
     request_queue_size = 2000
@@ -120,7 +138,7 @@ class WebTestServer(ThreadingMixIn, BaseHTTPServer.HTTPServer):
     daemon_threads = True
 
     def __init__(self, server_address, request_handler_cls,
-                 router, rewriter, bind_address,
+                 router, rewriter, bind_address, ws_doc_root=None,
                  config=None, use_ssl=False, key_file=None, certificate=None,
                  encrypt_after_connect=False, latency=None, http2=False, **kwargs):
         """Server for HTTP(s) Requests
@@ -145,6 +163,8 @@ class WebTestServer(ThreadingMixIn, BaseHTTPServer.HTTPServer):
 
         :param certificate: Path to certificate to use if SSL is enabled.
 
+        :param ws_doc_root: Document root for websockets
+
         :param encrypt_after_connect: For each connection, don't start encryption
                                       until a CONNECT message has been received.
                                       This enables the server to act as a
@@ -170,20 +190,21 @@ class WebTestServer(ThreadingMixIn, BaseHTTPServer.HTTPServer):
         else:
             hostname_port = ("",server_address[1])
 
-        #super doesn't work here because BaseHTTPServer.HTTPServer is old-style
-        BaseHTTPServer.HTTPServer.__init__(self, hostname_port, request_handler_cls, **kwargs)
+        http.server.HTTPServer.__init__(self, hostname_port, request_handler_cls, **kwargs)
 
         if config is not None:
             Server.config = config
         else:
             self.logger.debug("Using default configuration")
-            with ConfigBuilder(browser_host=server_address[0],
+            with ConfigBuilder(self.logger,
+                               browser_host=server_address[0],
                                ports={"http": [self.server_address[1]]}) as config:
                 assert config["ssl_config"] is None
                 Server.config = config
 
 
 
+        self.ws_doc_root = ws_doc_root
         self.key_file = key_file
         self.certificate = certificate
         self.encrypt_after_connect = use_ssl and encrypt_after_connect
@@ -205,22 +226,24 @@ class WebTestServer(ThreadingMixIn, BaseHTTPServer.HTTPServer):
     def handle_error(self, request, client_address):
         error = sys.exc_info()[1]
 
-        if ((isinstance(error, socket.error) and
+        if ((isinstance(error, OSError) and
              isinstance(error.args, tuple) and
              error.args[0] in self.acceptable_errors) or
             (isinstance(error, IOError) and
              error.errno in self.acceptable_errors)):
             pass  # remote hang up before the result is sent
         else:
-            self.logger.error(traceback.format_exc())
+            msg = traceback.format_exc()
+            self.logger.error(f"{type(error)} {error}")
+            self.logger.info(msg)
 
 
-class BaseWebTestRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
+class BaseWebTestRequestHandler(http.server.BaseHTTPRequestHandler):
     """RequestHandler for WebTestHttpd"""
 
     def __init__(self, *args, **kwargs):
         self.logger = get_logger()
-        BaseHTTPServer.BaseHTTPRequestHandler.__init__(self, *args, **kwargs)
+        http.server.BaseHTTPRequestHandler.__init__(self, *args, **kwargs)
 
     def finish_handling_h1(self, request_line_is_valid):
 
@@ -238,7 +261,7 @@ class BaseWebTestRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
             response.write()
             return
 
-        self.logger.debug("%s %s" % (request.method, request.request_path))
+        self.logger.debug(f"{request.method} {request.request_path}")
         handler = self.server.router.get_handler(request)
         self.finish_handling(request, response, handler)
 
@@ -265,7 +288,10 @@ class BaseWebTestRequestHandler(BaseHTTPServer.BaseHTTPRequestHandler):
             try:
                 handler(request, response)
             except HTTPException as e:
-                response.set_error(e.code, e.message)
+                if 500 <= e.code < 600:
+                    self.logger.warning("HTTPException in handler: %s" % e)
+                    self.logger.warning(traceback.format_exc())
+                response.set_error(e.code, str(e))
             except Exception as e:
                 self.respond_with_error(response, e)
         self.logger.debug("%i %s %s (%s) %i" % (response.status[0],
@@ -342,6 +368,14 @@ class Http2WebTestRequestHandler(BaseWebTestRequestHandler):
         self.logger.debug('(%s) Initiating h2 Connection' % self.uid)
 
         with self.conn as connection:
+            # Bootstrapping WebSockets with HTTP/2 specification requires
+            # ENABLE_CONNECT_PROTOCOL to be set in order to enable WebSocket
+            # over HTTP/2
+            new_settings = dict(connection.local_settings)
+            new_settings[SettingCodes.ENABLE_CONNECT_PROTOCOL] = 1
+            connection.local_settings.update(new_settings)
+            connection.local_settings.acknowledge()
+
             connection.initiate_connection()
             data = connection.data_to_send()
             window_size = connection.remote_settings.initial_window_size
@@ -380,21 +414,38 @@ class Http2WebTestRequestHandler(BaseWebTestRequestHandler):
                             stream_queues[frame.stream_id] = (self.start_stream_thread(frame, queue), queue)
                         stream_queues[frame.stream_id][1].put(frame)
 
-                        if isinstance(frame, StreamEnded) or (hasattr(frame, "stream_ended") and frame.stream_ended):
+                        if isinstance(frame, StreamEnded) or getattr(frame, "stream_ended", False):
                             del stream_queues[frame.stream_id]
 
-        except (socket.timeout, socket.error) as e:
-            self.logger.error('(%s) Closing Connection - \n%s' % (self.uid, str(e)))
+        except OSError as e:
+            self.logger.error(f'({self.uid}) Closing Connection - \n{str(e)}')
             if not self.close_connection:
                 self.close_connection = True
-                for stream_id, (thread, queue) in stream_queues.items():
-                    queue.put(None)
         except Exception as e:
-            self.logger.error('(%s) Unexpected Error - \n%s' % (self.uid, str(e)))
+            self.logger.error(f'({self.uid}) Unexpected Error - \n{str(e)}')
         finally:
             for stream_id, (thread, queue) in stream_queues.items():
+                queue.put(None)
                 thread.join()
 
+    def _is_extended_connect_frame(self, frame):
+        if not isinstance(frame, RequestReceived):
+            return False
+
+        method = extract_method_header(frame.headers)
+        if method != b"CONNECT":
+            return False
+
+        protocol = ""
+        for key, value in frame.headers:
+            if key in (b':protocol', ':protocol'):
+                protocol = isomorphic_encode(value)
+                break
+        if protocol != b"websocket":
+            raise ProtocolError(f"Invalid protocol {protocol} with CONNECT METHOD")
+
+        return True
+
     def start_stream_thread(self, frame, queue):
         """
         This starts a new thread to handle frames for a specific stream.
@@ -402,13 +453,111 @@ class Http2WebTestRequestHandler(BaseWebTestRequestHandler):
         :param queue: A queue object that the thread will use to check for new frames
         :return: The thread object that has already been started
         """
+        if self._is_extended_connect_frame(frame):
+            target = Http2WebTestRequestHandler._stream_ws_thread
+        else:
+            target = Http2WebTestRequestHandler._stream_thread
         t = threading.Thread(
-            target=Http2WebTestRequestHandler._stream_thread,
+            target=target,
             args=(self, frame.stream_id, queue)
         )
         t.start()
         return t
 
+    def _stream_ws_thread(self, stream_id, queue):
+        frame = queue.get(True, None)
+
+        if frame is None:
+            return
+
+        # Needs to be unbuffered for websockets.
+        rfile, wfile = os.pipe()
+        with os.fdopen(rfile, 'rb') as rfile, os.fdopen(wfile, 'wb', 0) as wfile:
+            stream_handler = H2HandlerCopy(self, frame, rfile)
+
+            h2request = H2Request(stream_handler)
+            h2response = H2Response(stream_handler, h2request)
+
+            dispatcher = dispatch.Dispatcher(self.server.ws_doc_root, None, False)
+            if not dispatcher.get_handler_suite(stream_handler.path):
+                h2response.set_error(404)
+                h2response.write()
+                return
+
+            request_wrapper = _WebSocketRequest(stream_handler, h2response)
+
+            handshaker = WsH2Handshaker(request_wrapper, dispatcher)
+            try:
+                handshaker.do_handshake()
+            except HandshakeException as e:
+                self.logger.info('Handshake failed for error: %s' % e)
+                h2response.set_error(e.status)
+                h2response.write()
+                return
+            except AbortedByUserException:
+                h2response.write()
+                return
+
+            # h2 Handshaker prepares the headers but does not send them down the
+            # wire. Flush the headers here.
+            try:
+                h2response.write_status_headers()
+            except StreamClosedError:
+                # work around https://github.com/web-platform-tests/wpt/issues/27786
+                # The stream was already closed.
+                return
+
+            request_wrapper._dispatcher = dispatcher
+
+            # we need two threads:
+            # - one to handle the frame queue
+            # - one to handle the request (dispatcher.transfer_data is blocking)
+            # the alternative is to have only one (blocking) thread. That thread
+            # will call transfer_data. That would require a special case in
+            # handle_one_request, to bypass the queue and write data to wfile
+            # directly.
+            t = threading.Thread(
+                target=Http2WebTestRequestHandler._stream_ws_sub_thread,
+                args=(self, request_wrapper, stream_handler, queue)
+            )
+            t.start()
+
+            while not self.close_connection:
+                try:
+                    frame = queue.get(True, 1)
+                except Empty:
+                    continue
+
+                if isinstance(frame, DataReceived):
+                    wfile.write(frame.data)
+                    if frame.stream_ended:
+                        raise NotImplementedError("frame.stream_ended")
+                elif frame is None or isinstance(frame, (StreamReset, StreamEnded, ConnectionTerminated)):
+                    self.logger.error(f'({self.uid} - {stream_id}) Stream Reset, Thread Closing')
+                    break
+
+        t.join()
+
+    def _stream_ws_sub_thread(self, request, stream_handler, queue):
+        dispatcher = request._dispatcher
+        try:
+            dispatcher.transfer_data(request)
+        except StreamClosedError:
+            # work around https://github.com/web-platform-tests/wpt/issues/27786
+            # The stream was already closed.
+            queue.put(None)
+            return
+
+        stream_id = stream_handler.h2_stream_id
+        with stream_handler.conn as connection:
+            try:
+                connection.end_stream(stream_id)
+                data = connection.data_to_send()
+                stream_handler.request.sendall(data)
+            except StreamClosedError:  # maybe the stream has already been closed
+                pass
+        queue.put(None)
+
     def _stream_thread(self, stream_id, queue):
         """
         This thread processes frames for a specific stream. It waits for frames to be placed
@@ -419,18 +568,38 @@ class Http2WebTestRequestHandler(BaseWebTestRequestHandler):
 
         # The file-like pipe object that will be used to share data to request object if data is received
         wfile = None
+        rfile = None
         request = None
         response = None
         req_handler = None
+
+        def cleanup():
+            # Try to close the files
+            # Ignore any exception (e.g. if the file handle was already closed for some reason).
+            if rfile:
+                try:
+                    rfile.close()
+                except OSError:
+                    pass
+            if wfile:
+                try:
+                    wfile.close()
+                except OSError:
+                    pass
+
         while not self.close_connection:
-            # Wait for next frame, blocking
-            frame = queue.get(True, None)
-
-            self.logger.debug('(%s - %s) %s' % (self.uid, stream_id, str(frame)))
+            try:
+                frame = queue.get(True, 1)
+            except Empty:
+                # Restart to check for close_connection
+                continue
 
+            self.logger.debug(f'({self.uid} - {stream_id}) {str(frame)}')
             if isinstance(frame, RequestReceived):
-                rfile, wfile = os.pipe()
-                rfile, wfile = os.fdopen(rfile, 'rb'), os.fdopen(wfile, 'wb')
+                cleanup()
+
+                pipe_rfile, pipe_wfile = os.pipe()
+                (rfile, wfile) = os.fdopen(pipe_rfile, 'rb'), os.fdopen(pipe_wfile, 'wb')
 
                 stream_handler = H2HandlerCopy(self, frame, rfile)
 
@@ -453,29 +622,35 @@ class Http2WebTestRequestHandler(BaseWebTestRequestHandler):
                 if hasattr(req_handler, 'handle_data'):
                     req_handler.handle_data(frame, request, response)
 
-                if frame.stream_ended:
-                    wfile.close()
             elif frame is None or isinstance(frame, (StreamReset, StreamEnded, ConnectionTerminated)):
-                self.logger.debug('(%s - %s) Stream Reset, Thread Closing' % (self.uid, stream_id))
+                self.logger.debug(f'({self.uid} - {stream_id}) Stream Reset, Thread Closing')
                 break
 
             if request is not None:
                 request.frames.append(frame)
 
-            if hasattr(frame, "stream_ended") and frame.stream_ended:
-                self.finish_handling(request, response, req_handler)
+            if getattr(frame, "stream_ended", False):
+                try:
+                    self.finish_handling(request, response, req_handler)
+                except StreamClosedError:
+                    self.logger.debug('(%s - %s) Unable to write response; stream closed' %
+                                    (self.uid, stream_id))
+                break
+
+        cleanup()
 
     def frame_handler(self, request, response, handler):
         try:
             return handler.frame_handler(request)
         except HTTPException as e:
-            response.set_error(e.code, e.message)
+            response.set_error(e.code, str(e))
             response.write()
         except Exception as e:
             self.respond_with_error(response, e)
             response.write()
 
-class H2ConnectionGuard(object):
+
+class H2ConnectionGuard:
     """H2Connection objects are not threadsafe, so this keeps thread safety"""
     lock = threading.Lock()
 
@@ -491,10 +666,12 @@ class H2ConnectionGuard(object):
         self.lock.release()
 
 
-class H2Headers(dict):
+class H2Headers(Dict[bytes, bytes]):
     def __init__(self, headers):
         self.raw_headers = OrderedDict()
         for key, val in headers:
+            key = isomorphic_decode(key)
+            val = isomorphic_decode(val)
             self.raw_headers[key] = val
             dict.__setitem__(self, self._convert_h2_header_to_h1(key), val)
 
@@ -509,7 +686,7 @@ class H2Headers(dict):
         return ['dummy function']
 
 
-class H2HandlerCopy(object):
+class H2HandlerCopy:
     def __init__(self, handler, req_frame, rfile):
         self.headers = H2Headers(req_frame.headers)
         self.command = self.headers['method']
@@ -517,6 +694,7 @@ class H2HandlerCopy(object):
         self.h2_stream_id = req_frame.stream_id
         self.server = handler.server
         self.protocol_version = handler.protocol_version
+        self.client_address = handler.client_address
         self.raw_requestline = ''
         self.rfile = rfile
         self.request = handler.request
@@ -553,12 +731,11 @@ class Http1WebTestRequestHandler(BaseWebTestRequestHandler):
             if response:
                 response.set_error(500, err)
                 response.write()
-            self.logger.error(err)
 
     def get_request_line(self):
         try:
             self.raw_requestline = self.rfile.readline(65537)
-        except socket.error:
+        except OSError:
             self.close_connection = True
             return False
         if len(self.raw_requestline) > 65536:
@@ -570,7 +747,7 @@ class Http1WebTestRequestHandler(BaseWebTestRequestHandler):
             self.close_connection = True
         return True
 
-class WebTestHttpd(object):
+class WebTestHttpd:
     """
     :param host: Host from which to serve (default: 127.0.0.1)
     :param port: Port from which to serve (default: 8000)
@@ -585,6 +762,7 @@ class WebTestHttpd(object):
                                   self-proxy.
     :param router_cls: Router class to use when matching URLs to handlers
     :param doc_root: Document root for serving files
+    :param ws_doc_root: Document root for websockets
     :param routes: List of routes with which to initialize the router
     :param rewriter_cls: Class to use for request rewriter
     :param rewrites: List of rewrites with which to initialize the rewriter_cls
@@ -627,7 +805,7 @@ class WebTestHttpd(object):
     def __init__(self, host="127.0.0.1", port=8000,
                  server_cls=None, handler_cls=Http1WebTestRequestHandler,
                  use_ssl=False, key_file=None, certificate=None, encrypt_after_connect=False,
-                 router_cls=Router, doc_root=os.curdir, routes=None,
+                 router_cls=Router, doc_root=os.curdir, ws_doc_root=None, routes=None,
                  rewriter_cls=RequestRewriter, bind_address=True, rewrites=None,
                  latency=None, config=None, http2=False):
 
@@ -648,9 +826,9 @@ class WebTestHttpd(object):
 
         if use_ssl:
             if not os.path.exists(key_file):
-                raise ValueError("SSL certificate not found: {}".format(key_file))
+                raise ValueError(f"SSL certificate not found: {key_file}")
             if not os.path.exists(certificate):
-                raise ValueError("SSL key not found: {}".format(certificate))
+                raise ValueError(f"SSL key not found: {certificate}")
 
         try:
             self.httpd = server_cls((host, port),
@@ -659,6 +837,7 @@ class WebTestHttpd(object):
                                     self.rewriter,
                                     config=config,
                                     bind_address=bind_address,
+                                    ws_doc_root=ws_doc_root,
                                     use_ssl=use_ssl,
                                     key_file=key_file,
                                     certificate=certificate,
@@ -669,24 +848,22 @@ class WebTestHttpd(object):
 
             _host, self.port = self.httpd.socket.getsockname()
         except Exception:
-            self.logger.error("Failed to start HTTP server. "
-                              "You may need to edit /etc/hosts or similar, see README.md.")
+            self.logger.critical("Failed to start HTTP server on port %s; "
+                                 "is something already using that port?" % port)
             raise
 
-    def start(self, block=False):
+    def start(self):
         """Start the server.
 
         :param block: True to run the server on the current thread, blocking,
                       False to run on a separate thread."""
         http_type = "http2" if self.http2 else "https" if self.use_ssl else "http"
-        self.logger.info("Starting %s server on %s:%s" % (http_type, self.host, self.port))
+        http_scheme = "https" if self.use_ssl else "http"
+        self.logger.info(f"Starting {http_type} server on {http_scheme}://{self.host}:{self.port}")
         self.started = True
-        if block:
-            self.httpd.serve_forever()
-        else:
-            self.server_thread = threading.Thread(target=self.httpd.serve_forever)
-            self.server_thread.setDaemon(True)  # don't hang on exit
-            self.server_thread.start()
+        self.server_thread = threading.Thread(target=self.httpd.serve_forever)
+        self.server_thread.daemon = True  # don't hang on exit
+        self.server_thread.start()
 
     def stop(self):
         """
@@ -700,7 +877,7 @@ class WebTestHttpd(object):
                 self.httpd.server_close()
                 self.server_thread.join()
                 self.server_thread = None
-                self.logger.info("Stopped http server on %s:%s" % (self.host, self.port))
+                self.logger.info(f"Stopped http server on {self.host}:{self.port}")
             except AttributeError:
                 pass
             self.started = False
@@ -711,5 +888,57 @@ class WebTestHttpd(object):
             return None
 
         return urlunsplit(("http" if not self.use_ssl else "https",
-                           "%s:%s" % (self.host, self.port),
+                           f"{self.host}:{self.port}",
                            path, query, fragment))
+
+
+class _WebSocketConnection:
+    def __init__(self, request_handler, response):
+        """Mimic mod_python mp_conn.
+
+        :param request_handler: A H2HandlerCopy instance.
+
+        :param response: A H2Response instance.
+        """
+
+        self._request_handler = request_handler
+        self._response = response
+
+        self.remote_addr = self._request_handler.client_address
+
+    def write(self, data):
+        self._response.writer.write_data(data, False)
+
+    def read(self, length):
+        return self._request_handler.rfile.read(length)
+
+
+class _WebSocketRequest:
+    def __init__(self, request_handler, response):
+        """Mimic mod_python request.
+
+        :param request_handler: A H2HandlerCopy instance.
+
+        :param response: A H2Response instance.
+        """
+
+        self.connection = _WebSocketConnection(request_handler, response)
+        self.protocol = "HTTP/2"
+        self._response = response
+
+        self.uri = request_handler.path
+        self.unparsed_uri = request_handler.path
+        self.method = request_handler.command
+        # read headers from request_handler
+        self.headers_in = request_handler.headers
+        # write headers directly into H2Response
+        self.headers_out = response.headers
+
+    # proxies status to H2Response
+    @property
+    def status(self):
+        return self._response.status
+
+    @status.setter
+    def status(self, status):
+        self._response.status = status
